@@ -1,5 +1,5 @@
 /*
-  Copyright 2012 David Robillard <http://drobilla.net>
+  Copyright 2012-2016 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -31,31 +31,43 @@
 
 #define USTR(str) ((const uint8_t*)(str))
 
+static const SerdStyle style = (SerdStyle)(
+	SERD_STYLE_ABBREVIATED|SERD_STYLE_RESOLVED|SERD_STYLE_CURIED);
+
 typedef enum {
-	MODE_NORMAL,
-	MODE_SEQUENCE,
+	MODE_SUBJECT,
+	MODE_BODY,
+	MODE_SEQUENCE
 } ReadMode;
 
 struct SratomImpl {
 	LV2_URID_Map*     map;
 	LV2_Atom_Forge    forge;
-	LV2_URID          atom_Event;
-	LV2_URID          midi_MidiEvent;
-	unsigned          next_id;
+	SerdEnv*          env;
 	SerdNode          base_uri;
+	SerdURI           base;
 	SerdStatementSink write_statement;
 	SerdEndSink       end_anon;
 	void*             handle;
-	bool              pretty_numbers;
+	LV2_URID          atom_Event;
+	LV2_URID          atom_frameTime;
+	LV2_URID          atom_beatTime;
+	LV2_URID          midi_MidiEvent;
+	unsigned          next_id;
+	SratomObjectMode  object_mode;
+	uint32_t          seq_unit;
 	struct {
 		SordNode* atom_childType;
 		SordNode* atom_frameTime;
+		SordNode* atom_beatTime;
 		SordNode* rdf_first;
 		SordNode* rdf_rest;
 		SordNode* rdf_type;
 		SordNode* rdf_value;
 		SordNode* xsd_base64Binary;
 	} nodes;
+
+	bool pretty_numbers;
 };
 
 static void
@@ -70,15 +82,16 @@ SRATOM_API
 Sratom*
 sratom_new(LV2_URID_Map* map)
 {
-	Sratom* sratom = (Sratom*)malloc(sizeof(Sratom));
-	sratom->map            = map;
-	sratom->atom_Event     = map->map(map->handle, LV2_ATOM__Event);
-	sratom->midi_MidiEvent = map->map(map->handle, LV2_MIDI__MidiEvent);
-	sratom->next_id        = 0;
-	sratom->base_uri       = SERD_NODE_NULL;
-	sratom->pretty_numbers = false;
-	memset(&sratom->nodes, 0, sizeof(sratom->nodes));
-	lv2_atom_forge_init(&sratom->forge, map);
+	Sratom* sratom = (Sratom*)calloc(1, sizeof(Sratom));
+	if (sratom) {
+		sratom->map            = map;
+		sratom->atom_Event     = map->map(map->handle, LV2_ATOM__Event);
+		sratom->atom_frameTime = map->map(map->handle, LV2_ATOM__frameTime);
+		sratom->atom_beatTime  = map->map(map->handle, LV2_ATOM__beatTime);
+		sratom->midi_MidiEvent = map->map(map->handle, LV2_MIDI__MidiEvent);
+		sratom->object_mode    = SRATOM_OBJECT_MODE_BLANK;
+		lv2_atom_forge_init(&sratom->forge, map);
+	}
 	return sratom;
 }
 
@@ -86,24 +99,33 @@ SRATOM_API
 void
 sratom_free(Sratom* sratom)
 {
+	serd_node_free(&sratom->base_uri);
 	free(sratom);
 }
 
+SRATOM_API
+void
+sratom_set_env(Sratom* sratom, SerdEnv* env)
+{
+	sratom->env = env;
+}
+
 SRATOM_API
 void
 sratom_set_sink(Sratom*           sratom,
                 const char*       base_uri,
-                SerdStatementSink write_statement,
-                SerdEndSink       end_anon,
+                SerdStatementSink sink,
+                SerdEndSink       end_sink,
                 void*             handle)
 {
 	if (base_uri) {
 		serd_node_free(&sratom->base_uri);
 		sratom->base_uri = serd_node_new_uri_from_string(
 			USTR(base_uri), NULL, NULL);
+		serd_uri_parse(sratom->base_uri.buf, &sratom->base);
 	}
-	sratom->write_statement = write_statement;
-	sratom->end_anon        = end_anon;
+	sratom->write_statement = sink;
+	sratom->end_anon        = end_sink;
 	sratom->handle          = handle;
 }
 
@@ -115,6 +137,14 @@ sratom_set_pretty_numbers(Sratom* sratom,
 	sratom->pretty_numbers = pretty_numbers;
 }
 
+SRATOM_API
+void
+sratom_set_object_mode(Sratom*          sratom,
+                       SratomObjectMode object_mode)
+{
+	sratom->object_mode = object_mode;
+}
+
 static void
 gensym(SerdNode* out, char c, unsigned num)
 {
@@ -131,7 +161,7 @@ list_append(Sratom*           sratom,
             SerdNode*         node,
             uint32_t          size,
             uint32_t          type,
-            void*             body)
+            const void*       body)
 {
 	// Generate a list node
 	gensym(node, 'l', sratom->next_id);
@@ -163,18 +193,25 @@ list_end(SerdStatementSink sink,
 
 static void
 start_object(Sratom*         sratom,
-             uint32_t        flags,
+             uint32_t*       flags,
              const SerdNode* subject,
              const SerdNode* predicate,
              const SerdNode* node,
              const char*     type)
 {
-	sratom->write_statement(sratom->handle, flags|SERD_ANON_O_BEGIN, NULL,
-	                        subject, predicate, node, NULL, NULL);
+	if (subject && predicate) {
+		sratom->write_statement(sratom->handle, *flags|SERD_ANON_O_BEGIN, NULL,
+		                        subject, predicate, node, NULL, NULL);
+		// Start abbreviating object properties
+		*flags |= SERD_ANON_CONT;
+
+		// Object is in a list, stop list abbreviating if necessary
+		*flags &= ~SERD_LIST_CONT;
+	}
 	if (type) {
 		SerdNode p = serd_node_from_string(SERD_URI, NS_RDF "type");
 		SerdNode o = serd_node_from_string(SERD_URI, USTR(type));
-		sratom->write_statement(sratom->handle, SERD_ANON_CONT, NULL,
+		sratom->write_statement(sratom->handle, *flags, NULL,
 		                        node, &p, &o, NULL, NULL);
 	}
 }
@@ -187,6 +224,22 @@ path_is_absolute(const char* path)
 	            && (path[2] == '/' || path[2] == '\\')));
 }
 
+static SerdNode
+number_type(const Sratom* sratom, const uint8_t* type)
+{
+	if (sratom->pretty_numbers &&
+	    (!strcmp((const char*)type, (const char*)NS_XSD "int") ||
+	     !strcmp((const char*)type, (const char*)NS_XSD "long"))) {
+		return serd_node_from_string(SERD_URI, NS_XSD "integer");
+	} else if (sratom->pretty_numbers &&
+	           (!strcmp((const char*)type, (const char*)NS_XSD "float") ||
+	            !strcmp((const char*)type, (const char*)NS_XSD "double"))) {
+		return serd_node_from_string(SERD_URI, NS_XSD "decimal");
+	} else {
+		return serd_node_from_string(SERD_URI, type);
+	}
+}
+
 SRATOM_API
 int
 sratom_write(Sratom*         sratom,
@@ -214,8 +267,9 @@ sratom_write(Sratom*         sratom,
 	} else if (type_urid == sratom->forge.Chunk) {
 		datatype = serd_node_from_string(SERD_URI, NS_XSD "base64Binary");
 		object   = serd_node_new_blob(body, size, true);
+		new_node = true;
 	} else if (type_urid == sratom->forge.Literal) {
-		LV2_Atom_Literal_Body* lit = (LV2_Atom_Literal_Body*)body;
+		const LV2_Atom_Literal_Body* lit = (const LV2_Atom_Literal_Body*)body;
 		const uint8_t*         str = USTR(lit + 1);
 		object = serd_node_from_string(SERD_LITERAL, str);
 		if (lit->datatype) {
@@ -229,20 +283,19 @@ sratom_write(Sratom*         sratom,
 				language = serd_node_from_string(
 					SERD_LITERAL, USTR(lang + prefix_len));
 			} else {
-				fprintf(stderr, "Unknown language URI <%s>\n", lang);
+				fprintf(stderr, "Unknown language URID %d\n", lit->lang);
 			}
 		}
 	} else if (type_urid == sratom->forge.URID) {
-		const uint32_t id  = *(const uint32_t*)body;
-		const uint8_t* str = USTR(unmap->unmap(unmap->handle, id));
+		const uint32_t urid = *(const uint32_t*)body;
+		const uint8_t* str  = USTR(unmap->unmap(unmap->handle, urid));
 		object = serd_node_from_string(SERD_URI, str);
 	} else if (type_urid == sratom->forge.Path) {
 		const uint8_t* str = USTR(body);
 		if (path_is_absolute((const char*)str)) {
 			new_node = true;
-			object   = serd_node_new_file_uri(str, NULL, NULL, false);
+			object   = serd_node_new_file_uri(str, NULL, NULL, true);
 		} else {
-			SerdURI base_uri = SERD_URI_NULL;
 			if (!sratom->base_uri.buf ||
 			    strncmp((const char*)sratom->base_uri.buf, "file://", 7)) {
 				fprintf(stderr, "warning: Relative path but base is not a file URI.\n");
@@ -250,12 +303,9 @@ sratom_write(Sratom*         sratom,
 				object   = serd_node_from_string(SERD_LITERAL, str);
 				datatype = serd_node_from_string(SERD_URI, USTR(LV2_ATOM__Path));
 			} else {
-				if (sratom->base_uri.buf) {
-					serd_uri_parse(sratom->base_uri.buf, &base_uri);
-				}
 				new_node = true;
-				SerdNode rel = serd_node_new_file_uri(str, NULL, NULL, false);
-				object = serd_node_new_uri_from_node(&rel, &base_uri, NULL);
+				SerdNode rel = serd_node_new_file_uri(str, NULL, NULL, true);
+				object = serd_node_new_uri_from_node(&rel, &sratom->base, NULL);
 				serd_node_free(&rel);
 			}
 		}
@@ -264,24 +314,20 @@ sratom_write(Sratom*         sratom,
 		object = serd_node_from_string(SERD_URI, str);
 	} else if (type_urid == sratom->forge.Int) {
 		new_node = true;
-		object   = serd_node_new_integer(*(int32_t*)body);
-		datatype = serd_node_from_string(SERD_URI, (sratom->pretty_numbers)
-		                                 ? NS_XSD "integer" : NS_XSD "int");
+		object   = serd_node_new_integer(*(const int32_t*)body);
+		datatype = number_type(sratom, NS_XSD "int");
 	} else if (type_urid == sratom->forge.Long) {
 		new_node = true;
-		object   = serd_node_new_integer(*(int64_t*)body);
-		datatype = serd_node_from_string(SERD_URI, (sratom->pretty_numbers)
-		                                 ? NS_XSD "integer" : NS_XSD "long");
+		object   = serd_node_new_integer(*(const int64_t*)body);
+		datatype = number_type(sratom, NS_XSD "long");
 	} else if (type_urid == sratom->forge.Float) {
 		new_node = true;
-		object   = serd_node_new_decimal(*(float*)body, 8);
-		datatype = serd_node_from_string(SERD_URI, (sratom->pretty_numbers)
-		                                 ? NS_XSD "decimal" : NS_XSD "float");
+		object   = serd_node_new_decimal(*(const float*)body, 8);
+		datatype = number_type(sratom, NS_XSD "float");
 	} else if (type_urid == sratom->forge.Double) {
 		new_node = true;
-		object   = serd_node_new_decimal(*(double*)body, 16);
-		datatype = serd_node_from_string(SERD_URI, (sratom->pretty_numbers)
-		                                 ? NS_XSD "decimal" : NS_XSD "double");
+		object   = serd_node_new_decimal(*(const double*)body, 16);
+		datatype = number_type(sratom, NS_XSD "double");
 	} else if (type_urid == sratom->forge.Bool) {
 		const int32_t val = *(const int32_t*)body;
 		datatype = serd_node_from_string(SERD_URI, NS_XSD "boolean");
@@ -293,20 +339,26 @@ sratom_write(Sratom*         sratom,
 		uint8_t* str = (uint8_t*)calloc(size * 2 + 1, 1);
 		for (uint32_t i = 0; i < size; ++i) {
 			snprintf((char*)str + (2 * i), size * 2 + 1, "%02X",
-			         (unsigned)(uint8_t)*((uint8_t*)body + i));
+			         (unsigned)(uint8_t)*((const uint8_t*)body + i));
 		}
 		object = serd_node_from_string(SERD_LITERAL, USTR(str));
 	} else if (type_urid == sratom->atom_Event) {
 		const LV2_Atom_Event* ev = (const LV2_Atom_Event*)body;
 		gensym(&id, 'e', sratom->next_id++);
-		start_object(sratom, flags, subject, predicate, &id, NULL);
-		// TODO: beat time
-		SerdNode time = serd_node_new_integer(ev->time.frames);
-		SerdNode p    = serd_node_from_string(SERD_URI,
-		                                      USTR(LV2_ATOM__frameTime));
-		datatype = serd_node_from_string(SERD_URI, NS_XSD "decimal");
+		start_object(sratom, &flags, subject, predicate, &id, NULL);
+		SerdNode time;
+		SerdNode p;
+		if (sratom->seq_unit == sratom->atom_beatTime) {
+			time     = serd_node_new_decimal(ev->time.beats, 16);
+			p        = serd_node_from_string(SERD_URI, USTR(LV2_ATOM__beatTime));
+			datatype = number_type(sratom, NS_XSD "double");
+		} else {
+			time     = serd_node_new_integer(ev->time.frames);
+			p        = serd_node_from_string(SERD_URI, USTR(LV2_ATOM__frameTime));
+			datatype = number_type(sratom, NS_XSD "long");
+		}
 		sratom->write_statement(sratom->handle, SERD_ANON_CONT, NULL,
-		     &id, &p, &time, &datatype, &language);
+		                        &id, &p, &time, &datatype, &language);
 		serd_node_free(&time);
 
 		p = serd_node_from_string(SERD_URI, NS_RDF "value");
@@ -317,7 +369,7 @@ sratom_write(Sratom*         sratom,
 		}
 	} else if (type_urid == sratom->forge.Tuple) {
 		gensym(&id, 't', sratom->next_id++);
-		start_object(sratom, flags, subject, predicate, &id, type);
+		start_object(sratom, &flags, subject, predicate, &id, type);
 		SerdNode p = serd_node_from_string(SERD_URI, NS_RDF "value");
 		flags |= SERD_LIST_O_BEGIN;
 		LV2_ATOM_TUPLE_BODY_FOREACH(body, size, i) {
@@ -331,15 +383,15 @@ sratom_write(Sratom*         sratom,
 	} else if (type_urid == sratom->forge.Vector) {
 		const LV2_Atom_Vector_Body* vec  = (const LV2_Atom_Vector_Body*)body;
 		gensym(&id, 'v', sratom->next_id++);
-		start_object(sratom, flags, subject, predicate, &id, type);
+		start_object(sratom, &flags, subject, predicate, &id, type);
 		SerdNode p = serd_node_from_string(SERD_URI, (const uint8_t*)LV2_ATOM__childType);
 		SerdNode child_type = serd_node_from_string(
 			SERD_URI, (const uint8_t*)unmap->unmap(unmap->handle, vec->child_type));
 		sratom->write_statement(sratom->handle, flags, NULL, &id, &p, &child_type, NULL, NULL);
 		p = serd_node_from_string(SERD_URI, NS_RDF "value");
 		flags |= SERD_LIST_O_BEGIN;
-		for (char* i = (char*)(vec + 1);
-		     i < (char*)vec + size;
+		for (const char* i = (const char*)(vec + 1);
+		     i < (const char*)vec + size;
 		     i += vec->child_size) {
 			list_append(sratom, unmap, &flags, &id, &p, &node,
 			            vec->child_size, vec->child_type, i);
@@ -348,54 +400,71 @@ sratom_write(Sratom*         sratom,
 		if (sratom->end_anon) {
 			sratom->end_anon(sratom->handle, &id);
 		}
-	} else if (type_urid == sratom->forge.Blank) {
+	} else if (lv2_atom_forge_is_object_type(&sratom->forge, type_urid)) {
 		const LV2_Atom_Object_Body* obj   = (const LV2_Atom_Object_Body*)body;
 		const char*                 otype = unmap->unmap(unmap->handle,
 		                                                 obj->otype);
-		gensym(&id, 'b', sratom->next_id++);
-		start_object(sratom, flags, subject, predicate, &id, otype);
+
+		if (lv2_atom_forge_is_blank(&sratom->forge, type_urid, obj)) {
+			gensym(&id, 'b', sratom->next_id++);
+			start_object(sratom, &flags, subject, predicate, &id, otype);
+		} else {
+			id = serd_node_from_string(
+				SERD_URI, (const uint8_t*)unmap->unmap(unmap->handle, obj->id));
+			flags = 0;
+			start_object(sratom, &flags, NULL, NULL, &id, otype);
+		}
 		LV2_ATOM_OBJECT_BODY_FOREACH(obj, size, prop) {
 			const char* const key  = unmap->unmap(unmap->handle, prop->key);
 			SerdNode          pred = serd_node_from_string(SERD_URI, USTR(key));
-			sratom_write(sratom, unmap, flags|SERD_ANON_CONT, &id, &pred,
+			sratom_write(sratom, unmap, flags, &id, &pred,
 			             prop->value.type, prop->value.size,
 			             LV2_ATOM_BODY(&prop->value));
 		}
-		if (sratom->end_anon) {
+		if (sratom->end_anon && (flags & SERD_ANON_CONT)) {
 			sratom->end_anon(sratom->handle, &id);
 		}
 	} else if (type_urid == sratom->forge.Sequence) {
 		const LV2_Atom_Sequence_Body* seq = (const LV2_Atom_Sequence_Body*)body;
 		gensym(&id, 'v', sratom->next_id++);
-		start_object(sratom, flags, subject, predicate, &id, type);
+		start_object(sratom, &flags, subject, predicate, &id, type);
 		SerdNode p = serd_node_from_string(SERD_URI, NS_RDF "value");
 		flags |= SERD_LIST_O_BEGIN;
 		LV2_ATOM_SEQUENCE_BODY_FOREACH(seq, size, ev) {
+			sratom->seq_unit = seq->unit;
 			list_append(sratom, unmap, &flags, &id, &p, &node,
 			            sizeof(LV2_Atom_Event) + ev->body.size,
 			            sratom->atom_Event,
 			            ev);
 		}
 		list_end(sratom->write_statement, sratom->handle, &flags, &id, &p);
-		if (sratom->end_anon) {
+		if (sratom->end_anon && subject && predicate) {
 			sratom->end_anon(sratom->handle, &id);
 		}
 	} else {
 		gensym(&id, 'b', sratom->next_id++);
-		start_object(sratom, flags, subject, predicate, &id, type);
+		start_object(sratom, &flags, subject, predicate, &id, type);
 		SerdNode p = serd_node_from_string(SERD_URI, NS_RDF "value");
 		SerdNode o = serd_node_new_blob(body, size, true);
 		datatype = serd_node_from_string(SERD_URI, NS_XSD "base64Binary");
 		sratom->write_statement(sratom->handle, flags, NULL, &id, &p, &o, &datatype, NULL);
-		if (sratom->end_anon) {
+		if (sratom->end_anon && subject && predicate) {
 			sratom->end_anon(sratom->handle, &id);
 		}
 		serd_node_free(&o);
 	}
 
 	if (object.buf) {
+		SerdNode def_s = serd_node_from_string(SERD_BLANK, USTR("atom"));
+		SerdNode def_p = serd_node_from_string(SERD_URI, USTR(NS_RDF "value"));
+		if (!subject) {
+			subject = &def_s;
+		}
+		if (!predicate) {
+			predicate = &def_p;
+		}
 		sratom->write_statement(sratom->handle, flags, NULL,
-		     subject, predicate, &object, &datatype, &language);
+		                        subject, predicate, &object, &datatype, &language);
 	}
 
 	if (new_node) {
@@ -416,32 +485,14 @@ sratom_to_turtle(Sratom*         sratom,
                  uint32_t        size,
                  const void*     body)
 {
-	SerdURI   buri = SERD_URI_NULL;
-	SerdNode  base = serd_node_new_uri_from_string(USTR(base_uri), NULL, &buri);
-	SerdEnv*  env  = serd_env_new(&base);
-	SerdChunk str  = { NULL, 0 };
-
-	serd_env_set_prefix_from_strings(env, USTR("midi"),
-	                                 USTR(LV2_MIDI_PREFIX));
-	serd_env_set_prefix_from_strings(env, USTR("atom"),
-	                                 USTR(LV2_ATOM_URI "#"));
-	serd_env_set_prefix_from_strings(env, USTR("rdf"), NS_RDF);
-	serd_env_set_prefix_from_strings(env, USTR("xsd"), NS_XSD);
-	serd_env_set_prefix_from_strings(env, USTR("eg"),
-	                                 USTR("http://example.org/"));
-
+	SerdURI     buri   = SERD_URI_NULL;
+	SerdNode    base   = serd_node_new_uri_from_string(USTR(base_uri), &sratom->base, &buri);
+	SerdEnv*    env    = sratom->env ? sratom->env : serd_env_new(NULL);
+	SerdChunk   str    = { NULL, 0 };
 	SerdWriter* writer = serd_writer_new(
-		SERD_TURTLE,
-		(SerdStyle)(SERD_STYLE_ABBREVIATED |
-		            SERD_STYLE_RESOLVED |
-		            SERD_STYLE_CURIED),
-		env, &buri, serd_chunk_sink, &str);
-
-	// Write @prefix directives
-	serd_env_foreach(env,
-	                 (SerdPrefixSink)serd_writer_set_prefix,
-	                 writer);
+		SERD_TURTLE, style, env, &buri, serd_chunk_sink, &str);
 
+	serd_env_set_base_uri(env, &base);
 	sratom_set_sink(sratom, base_uri,
 	                (SerdStatementSink)serd_writer_write_statement,
 	                (SerdEndSink)serd_writer_end_anon,
@@ -451,25 +502,11 @@ sratom_to_turtle(Sratom*         sratom,
 	serd_writer_finish(writer);
 
 	serd_writer_free(writer);
-	serd_env_free(env);
-	return (char*)serd_chunk_sink_finish(&str);
-}
-
-static const SordNode*
-get_object(SordModel*      model,
-           const SordNode* subject,
-           const SordNode* predicate)
-{
-	const SordNode* object = NULL;
-	SordQuad        q      = { subject, predicate, 0, 0 };
-	SordIter*       i      = sord_find(model, q);
-	if (!sord_iter_end(i)) {
-		SordQuad quad;
-		sord_iter_get(i, quad);
-		object = quad[SORD_OBJECT];
+	if (!sratom->env) {
+		serd_env_free(env);
 	}
-	sord_iter_free(i);
-	return object;
+	serd_node_free(&base);
+	return (char*)serd_chunk_sink_finish(&str);
 }
 
 static void
@@ -480,12 +517,42 @@ read_list_value(Sratom*         sratom,
                 const SordNode* node,
                 ReadMode        mode)
 {
-	const SordNode* first = get_object(model, node, sratom->nodes.rdf_first);
-	const SordNode* rest  = get_object(model, node, sratom->nodes.rdf_rest);
-	if (first && rest) {
-		read_node(sratom, forge, world, model, first, mode);
-		read_list_value(sratom, forge, world, model, rest, mode);
+	SordNode* fst = sord_get(model, node, sratom->nodes.rdf_first, NULL, NULL);
+	SordNode* rst = sord_get(model, node, sratom->nodes.rdf_rest, NULL, NULL);
+	if (fst && rst) {
+		read_node(sratom, forge, world, model, fst, mode);
+		read_list_value(sratom, forge, world, model, rst, mode);
 	}
+	sord_node_free(world, rst);
+	sord_node_free(world, fst);
+}
+
+static void
+read_resource(Sratom*         sratom,
+              LV2_Atom_Forge* forge,
+              SordWorld*      world,
+              SordModel*      model,
+              const SordNode* node,
+              LV2_URID        otype)
+{
+	LV2_URID_Map* map = sratom->map;
+	SordQuad      q   = { node, NULL, NULL, NULL };
+	SordIter*     i   = sord_find(model, q);
+	SordQuad      match;
+	for (; !sord_iter_end(i); sord_iter_next(i)) {
+		sord_iter_get(i, match);
+		const SordNode* p      = match[SORD_PREDICATE];
+		const SordNode* o      = match[SORD_OBJECT];
+		const char*     p_uri  = (const char*)sord_node_get_string(p);
+		uint32_t        p_urid = map->map(map->handle, p_uri);
+		if (!(sord_node_equals(p, sratom->nodes.rdf_type) &&
+		      sord_node_get_type(o) == SORD_URI &&
+		      map->map(map->handle, (const char*)sord_node_get_string(o)) == otype)) {
+			lv2_atom_forge_key(forge, p_urid);
+			read_node(sratom, forge, world, model, o, MODE_BODY);
+		}
+	}
+	sord_iter_free(i);
 }
 
 static uint32_t
@@ -503,9 +570,8 @@ atom_size(Sratom* sratom, uint32_t type_urid)
 		return sizeof(int32_t);
 	} else if (type_urid == sratom->forge.URID) {
 		return sizeof(uint32_t);
-	} else {
-		return 0;
 	}
+	return 0;
 }
 
 static void
@@ -520,24 +586,23 @@ read_node(Sratom*         sratom,
 	size_t        len = 0;
 	const char*   str = (const char*)sord_node_get_string_counted(node, &len);
 	if (sord_node_get_type(node) == SORD_LITERAL) {
-		char*       endptr;
 		SordNode*   datatype = sord_node_get_datatype(node);
 		const char* language = sord_node_get_language(node);
 		if (datatype) {
 			const char* type_uri = (const char*)sord_node_get_string(datatype);
-			if (!strcmp(type_uri, (char*)NS_XSD "int") ||
-			    !strcmp(type_uri, (char*)NS_XSD "integer")) {
-				lv2_atom_forge_int(forge, strtol(str, &endptr, 10));
-			} else if (!strcmp(type_uri, (char*)NS_XSD "long")) {
-				lv2_atom_forge_long(forge, strtol(str, &endptr, 10));
-			} else if (!strcmp(type_uri, (char*)NS_XSD "float") ||
-			           !strcmp(type_uri, (char*)NS_XSD "decimal")) {
-				lv2_atom_forge_float(forge, serd_strtod(str, &endptr));
-			} else if (!strcmp(type_uri, (char*)NS_XSD "double")) {
-				lv2_atom_forge_double(forge, serd_strtod(str, &endptr));
-			} else if (!strcmp(type_uri, (char*)NS_XSD "boolean")) {
+			if (!strcmp(type_uri, (const char*)NS_XSD "int") ||
+			    !strcmp(type_uri, (const char*)NS_XSD "integer")) {
+				lv2_atom_forge_int(forge, strtol(str, NULL, 10));
+			} else if (!strcmp(type_uri, (const char*)NS_XSD "long")) {
+				lv2_atom_forge_long(forge, strtol(str, NULL, 10));
+			} else if (!strcmp(type_uri, (const char*)NS_XSD "float") ||
+			           !strcmp(type_uri, (const char*)NS_XSD "decimal")) {
+				lv2_atom_forge_float(forge, serd_strtod(str, NULL));
+			} else if (!strcmp(type_uri, (const char*)NS_XSD "double")) {
+				lv2_atom_forge_double(forge, serd_strtod(str, NULL));
+			} else if (!strcmp(type_uri, (const char*)NS_XSD "boolean")) {
 				lv2_atom_forge_bool(forge, !strcmp(str, "true"));
-			} else if (!strcmp(type_uri, (char*)NS_XSD "base64Binary")) {
+			} else if (!strcmp(type_uri, (const char*)NS_XSD "base64Binary")) {
 				size_t size = 0;
 				void*  body = serd_base64_decode(USTR(str), len, &size);
 				lv2_atom_forge_atom(forge, size, forge->Chunk);
@@ -550,7 +615,6 @@ read_node(Sratom*         sratom,
 				for (const char* s = str; s < str + len; s += 2) {
 					unsigned num;
 					sscanf(s, "%2X", &num);
-					assert(num < UINT8_MAX);
 					const uint8_t c = num;
 					lv2_atom_forge_raw(forge, &c, 1);
 				}
@@ -573,19 +637,28 @@ read_node(Sratom*         sratom,
 		} else {
 			lv2_atom_forge_string(forge, str, len);
 		}
-	} else if (sord_node_get_type(node) == SORD_URI) {
+	} else if (sord_node_get_type(node) == SORD_URI &&
+	           !(sratom->object_mode == SRATOM_OBJECT_MODE_BLANK_SUBJECT
+	             && mode == MODE_SUBJECT)) {
 		if (!strcmp(str, (const char*)NS_RDF "nil")) {
 			lv2_atom_forge_atom(forge, 0, 0);
 		} else if (!strncmp(str, "file://", 7)) {
-			uint8_t* path = serd_file_uri_parse((const uint8_t*)str, NULL);
+			SerdURI uri;
+			serd_uri_parse((const uint8_t*)str, &uri);
+
+			SerdNode rel  = serd_node_new_relative_uri(&uri, &sratom->base, NULL, NULL);
+			uint8_t* path = serd_file_uri_parse(rel.buf, NULL);
 			lv2_atom_forge_path(forge, (const char*)path, strlen((const char*)path));
-			free(path);
+			serd_free(path);
+			serd_node_free(&rel);
 		} else {
 			lv2_atom_forge_urid(forge, map->map(map->handle, str));
 		}
 	} else {
-		const SordNode* type  = get_object(model, node, sratom->nodes.rdf_type);
-		const SordNode* value = get_object(model, node, sratom->nodes.rdf_value);
+		SordNode* type = sord_get(
+			model, node, sratom->nodes.rdf_type, NULL, NULL);
+		SordNode* value = sord_get(
+			model, node, sratom->nodes.rdf_value, NULL, NULL);
 
 		const uint8_t* type_uri  = NULL;
 		uint32_t       type_urid = 0;
@@ -596,29 +669,49 @@ read_node(Sratom*         sratom,
 
 		LV2_Atom_Forge_Frame frame = { 0, 0 };
 		if (mode == MODE_SEQUENCE) {
-			const SordNode* frame_time = get_object(
-				model, node, sratom->nodes.atom_frameTime);
-			const char* frame_time_str = frame_time
-				? (const char*)sord_node_get_string(frame_time)
-				: "";
-			lv2_atom_forge_frame_time(forge, serd_strtod(frame_time_str, NULL));
-			read_node(sratom, forge, world, model, value, MODE_NORMAL);
+			SordNode* time = sord_get(
+				model, node, sratom->nodes.atom_beatTime, NULL, NULL);
+			uint32_t seq_unit;
+			if (time) {
+				const char* time_str = (const char*)sord_node_get_string(time);
+				lv2_atom_forge_beat_time(forge, serd_strtod(time_str, NULL));
+				seq_unit = sratom->atom_beatTime;
+			} else {
+				time = sord_get(model, node, sratom->nodes.atom_frameTime, NULL, NULL);
+				const char* time_str = time
+					? (const char*)sord_node_get_string(time)
+					: "";
+				lv2_atom_forge_frame_time(forge, serd_strtod(time_str, NULL));
+				seq_unit = sratom->atom_frameTime;
+			}
+			read_node(sratom, forge, world, model, value, MODE_BODY);
+			sord_node_free(world, time);
+			sratom->seq_unit = seq_unit;
 		} else if (type_urid == sratom->forge.Tuple) {
 			lv2_atom_forge_tuple(forge, &frame);
-			read_list_value(sratom, forge, world, model, value, MODE_NORMAL);
+			read_list_value(sratom, forge, world, model, value, MODE_BODY);
 		} else if (type_urid == sratom->forge.Sequence) {
-			lv2_atom_forge_sequence_head(forge, &frame, 0);
+			const LV2_Atom_Forge_Ref ref = lv2_atom_forge_sequence_head(forge, &frame, 0);
+			sratom->seq_unit = 0;
 			read_list_value(sratom, forge, world, model, value, MODE_SEQUENCE);
+
+			LV2_Atom_Sequence* seq = (LV2_Atom_Sequence*)lv2_atom_forge_deref(forge, ref);
+			seq->body.unit = (sratom->seq_unit == sratom->atom_frameTime) ? 0 : sratom->seq_unit;
 		} else if (type_urid == sratom->forge.Vector) {
-			const SordNode* child_type_node = get_object(
-				model, node, sratom->nodes.atom_childType);
+			SordNode* child_type_node = sord_get(
+				model, node, sratom->nodes.atom_childType, NULL, NULL);
 			uint32_t child_type = map->map(
 				map->handle, (const char*)sord_node_get_string(child_type_node));
 			uint32_t child_size = atom_size(sratom, child_type);
 			if (child_size > 0) {
-				lv2_atom_forge_vector_head(forge, &frame, child_size, child_type);
-				read_list_value(sratom, forge, world, model, value, MODE_NORMAL);
+				LV2_Atom_Forge_Ref ref = lv2_atom_forge_vector_head(
+					forge, &frame, child_size, child_type);
+				read_list_value(sratom, forge, world, model, value, MODE_BODY);
+				lv2_atom_forge_pop(forge, &frame);
+				frame.ref = 0;
+				lv2_atom_forge_pad(forge, lv2_atom_forge_deref(forge, ref)->size);
 			}
+			sord_node_free(world, child_type_node);
 		} else if (value && sord_node_equals(sord_node_get_datatype(value),
 		                                     sratom->nodes.xsd_base64Binary)) {
 			size_t         vlen = 0;
@@ -628,29 +721,20 @@ read_node(Sratom*         sratom,
 			lv2_atom_forge_atom(forge, size, type_urid);
 			lv2_atom_forge_write(forge, body, size);
 			free(body);
+		} else if (sord_node_get_type(node) == SORD_URI) {
+			lv2_atom_forge_object(
+				forge, &frame, map->map(map->handle, str), type_urid);
+			read_resource(sratom, forge, world, model, node, type_urid);
 		} else {
-			lv2_atom_forge_blank(forge, &frame, sratom->next_id++, type_urid);
-			SordQuad match;
-
-			SordQuad q2 = { node, 0, 0, 0 };
-			SordIter* i = sord_find(model, q2);
-			for (;!sord_iter_end(i); sord_iter_next(i)) {
-				sord_iter_get(i, match);
-				const SordNode* p      = match[SORD_PREDICATE];
-				const char*     p_uri  = (const char*)sord_node_get_string(p);
-				uint32_t        p_urid = map->map(map->handle, p_uri);
-				if (!sord_node_equals(p, sratom->nodes.rdf_type)) {
-					// TODO: This will lose multiple rdf:type properties
-					lv2_atom_forge_property_head(forge, p_urid, 0);
-					read_node(sratom, forge, world, model,
-					          match[SORD_OBJECT], MODE_NORMAL);
-				}
-			}
-			sord_iter_free(i);
+			lv2_atom_forge_object(forge, &frame, 0, type_urid);
+			read_resource(sratom, forge, world, model, node, type_urid);
 		}
+
 		if (frame.ref) {
 			lv2_atom_forge_pop(forge, &frame);
 		}
+		sord_node_free(world, value);
+		sord_node_free(world, type);
 	}
 }
 
@@ -664,6 +748,7 @@ sratom_read(Sratom*         sratom,
 {
 	sratom->nodes.atom_childType   = sord_new_uri(world, USTR(LV2_ATOM__childType));
 	sratom->nodes.atom_frameTime   = sord_new_uri(world, USTR(LV2_ATOM__frameTime));
+	sratom->nodes.atom_beatTime    = sord_new_uri(world, USTR(LV2_ATOM__beatTime));
 	sratom->nodes.rdf_first        = sord_new_uri(world, NS_RDF "first");
 	sratom->nodes.rdf_rest         = sord_new_uri(world, NS_RDF "rest");
 	sratom->nodes.rdf_type         = sord_new_uri(world, NS_RDF "type");
@@ -671,7 +756,7 @@ sratom_read(Sratom*         sratom,
 	sratom->nodes.xsd_base64Binary = sord_new_uri(world, NS_XSD "base64Binary");
 
 	sratom->next_id = 1;
-	read_node(sratom, forge, world, model, node, MODE_NORMAL);
+	read_node(sratom, forge, world, model, node, MODE_SUBJECT);
 
 	sord_node_free(world, sratom->nodes.xsd_base64Binary);
 	sord_node_free(world, sratom->nodes.rdf_value);
@@ -679,6 +764,7 @@ sratom_read(Sratom*         sratom,
 	sord_node_free(world, sratom->nodes.rdf_rest);
 	sord_node_free(world, sratom->nodes.rdf_first);
 	sord_node_free(world, sratom->nodes.atom_frameTime);
+	sord_node_free(world, sratom->nodes.atom_beatTime);
 	sord_node_free(world, sratom->nodes.atom_childType);
 	memset(&sratom->nodes, 0, sizeof(sratom->nodes));
 }
@@ -712,35 +798,39 @@ sratom_from_turtle(Sratom*         sratom,
                    const char*     str)
 {
 	SerdChunk   out    = { NULL, 0 };
-	SerdNode    base   = serd_node_new_uri_from_string(USTR(base_uri), NULL, NULL);
+	SerdNode    base   = serd_node_new_uri_from_string(USTR(base_uri), &sratom->base, NULL);
 	SordWorld*  world  = sord_world_new();
 	SordModel*  model  = sord_new(world, SORD_SPO, false);
-	SerdEnv*    env    = serd_env_new(&base);
+	SerdEnv*    env    = sratom->env ? sratom->env : serd_env_new(&base);
 	SerdReader* reader = sord_new_reader(model, env, SERD_TURTLE, NULL);
 
 	if (!serd_reader_read_string(reader, (const uint8_t*)str)) {
 		SordNode* s = sord_node_from_serd_node(world, env, subject, 0, 0);
-		SordNode* p = sord_node_from_serd_node(world, env, predicate, 0, 0);
-		SordQuad  q = { s, p, 0, 0 };
-		SordIter* i = sord_find(model, q);
-		if (!sord_iter_end(i)) {
-			SordQuad result;
-			sord_iter_get(i, result);
-			lv2_atom_forge_set_sink(
-				&sratom->forge, sratom_forge_sink, sratom_forge_deref, &out);
-			sratom_read(sratom, &sratom->forge, world, model, result[SORD_OBJECT]);
+		lv2_atom_forge_set_sink(
+			&sratom->forge, sratom_forge_sink, sratom_forge_deref, &out);
+		if (subject && predicate) {
+			SordNode* p = sord_node_from_serd_node(world, env, predicate, 0, 0);
+			SordNode* o = sord_get(model, s, p, NULL, NULL);
+			if (o) {
+				sratom_read(sratom, &sratom->forge, world, model, o);
+				sord_node_free(world, o);
+			} else {
+				fprintf(stderr, "Failed to find node\n");
+			}
 		} else {
-			fprintf(stderr, "Failed to find node\n");
+			sratom_read(sratom, &sratom->forge, world, model, s);
 		}
-		sord_iter_free(i);
 	} else {
 		fprintf(stderr, "Failed to read Turtle\n");
 	}
 
 	serd_reader_free(reader);
-	serd_env_free(env);
+	if (!sratom->env) {
+		serd_env_free(env);
+	}
 	sord_free(model);
 	sord_world_free(world);
+	serd_node_free(&base);
 
 	return (LV2_Atom*)out.buf;
 }
